/*----------------------------------------------------------------------------
 Copyright:      Radig Ulrich  mailto: mail@ulrichradig.de
 Author:         Radig Ulrich
 Remarks:        
 known Problems: none
 Version:        24.10.2007
 Description:    Webserver Applikation
----------------------------------------------------------------------------*/

#include "httpd.h"
#include "webpage.h"


struct http_table http_entry[MAX_TCP_ENTRY];

//Hier wird das codierte Passwort aus config.h gespeichert.
unsigned char http_auth_passwort[20];

//----------------------------------------------------------------------------
//Variablenarry zum einfgen in Webseite %VA@00 bis %VA@09
unsigned int var_array[MAX_VAR_ARRAY] = {10,20,30,40,50,60,70,90,90,300};
//----------------------------------------------------------------------------

PROGMEM char http_header1[]={	"HTTP/1.0 200 Document follows\r\n"
								"Server: AVR_Small_Webserver\r\n"
								"Content-Type: text/html\r\n\r\n"};//LEN 87

PROGMEM char http_header2[]={	"HTTP/1.0 200 Document follows\r\n"
								"Server: AVR_Small_Webserver\r\n"
								"Content-Type: image/jpg\r\n\r\n"};//LEN 87

PROGMEM char http_header3[]={	"HTTP/1.0 401 Unauthorized\r\n"
								"Server: AVR_Small_Webserver\r\n"
								"WWW-Authenticate: Basic realm=\"NeedPassword\""
								"\r\nContent-Type: text/html\r\n\r\n"};//LEN 129

//----------------------------------------------------------------------------
//Kein Zugriff Seite bei keinem Passwort
PROGMEM char Page0[] = {"401 Unauthorized%END"};

unsigned char rx_header_end[5] = {"\r\n\r\n\0"};

//----------------------------------------------------------------------------
//Initialisierung des Httpd Testservers
void httpd_init (void)
{
	//Port fr die schaltbaren Ausgnge
	//Port kann nur bei ENC28J60 benutzt werden 
	//RTL8019 bentigt PORTA
#if USE_ENC28J60
	DDRA = 0x07; 
#endif

	//mal sehen was frei ist
#if USE_RTL8019
#endif

	//HTTP_AUTH_STRING 
	decode_base64((unsigned char*)HTTP_AUTH_STRING,http_auth_passwort);

	//Serverport und Anwendung eintragen
	add_tcp_app (80, (void(*)(unsigned char))httpd);
}
   
//----------------------------------------------------------------------------
//http Testserver
void httpd (unsigned char index)
{
	//Allererste Aufruf des Ports fr diese Anwendung
	//HTTPD_Anwendungsstack lschen
	if(tcp_entry[index].app_status==1)
	{
		httpd_stack_clear(index);
	}
	
	//HTTP wurde bei dieser Verbindung zum ersten mal aufgerufen oder
	//HTML Header Retransmission!
	if (tcp_entry[index].app_status <= 2)
	{	
		httpd_header_check (index);
	}
	
	//Der Header wurde gesendet und mit ACK besttigt (tcp_entry[index].app_status+1)
	//war das HTML Packet fertig, oder mssen weitere Daten gesendet werden, oder Retransmission?
	if (tcp_entry[index].app_status > 2 && tcp_entry[index].app_status < 0xFFFE)
	{
		httpd_data_send (index);
	}
	
	//Verbindung kann geschlossen werden! Alle HTML Daten wurden gesendet TCP Port kann
	//geschlossen werden (tcp_entry[index].app_status >= 0xFFFE)!!
	if (tcp_entry[index].app_status >= 0xFFFE)
	{
		tcp_entry[index].app_status = 0xFFFE;
		tcp_Port_close(index);
		return;
	}
	
	return;
}

//----------------------------------------------------------------------------
//HTTPD_STACK lschen
void httpd_stack_clear (unsigned char index)
{
	http_entry[index].http_header_type =0;
	http_entry[index].first_switch = 0;
	http_entry[index].http_auth = HTTP_AUTH_DEFAULT;
	http_entry[index].new_page_pointer = 0;
	http_entry[index].old_page_pointer = 0;
	http_entry[index].auth_ptr = http_auth_passwort;
	http_entry[index].hdr_end_pointer = rx_header_end;
	HTTP_DEBUG("\r\n**** NEUE HTTP ANFORDERUNG ****\r\n\r\n");	
	return;
}

//----------------------------------------------------------------------------
//Eintreffenden Header vom Client checken
void httpd_header_check (unsigned char index)
{
	//finden der Authorization und das Ende im Header auch ber mehrere Packete hinweg!!	
	if(*http_entry[index].hdr_end_pointer != 0)
	{		
		for(unsigned int a=TCP_DATA_START_VAR;a<(TCP_DATA_END_VAR);a++)
		{	
			HTTP_DEBUG("%c",eth_buffer[a]);
			
			if(!http_entry[index].http_auth) 
			{
				if (eth_buffer[a] != *http_entry[index].auth_ptr++)
				{
					http_entry[index].auth_ptr = http_auth_passwort;
				}
				if(*http_entry[index].auth_ptr == 0) 
				{
					http_entry[index].http_auth = 1;
					HTTP_DEBUG("  <---LOGIN OK!--->\r\n");
				}
			}
			
			if (eth_buffer[a] != *http_entry[index].hdr_end_pointer++)
			{
				http_entry[index].hdr_end_pointer = rx_header_end;
			}
			
			//Das Headerende wird mit (CR+LF+CR+LF) angezeigt!
			if(*http_entry[index].hdr_end_pointer == 0) 
			{
				HTTP_DEBUG("<---HEADER ENDE ERREICHT!--->\r\n");
				break;
			}
		}
	}
	
	//Standart INDEX.HTM Seite
	if (!http_entry[index].http_header_type)
	{
		http_entry[index].new_page_pointer =  Page1;
	}
	//Standart Bild von der Rakete
	if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"raket.gif\0")!=0) 
	{
		http_entry[index].http_header_type = 1;
		http_entry[index].new_page_pointer =  Page2;
	}
	//Standart Bild vom Balken Seite
	if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"balke.gif\0")!=0) 
	{
		http_entry[index].http_header_type = 1;
		http_entry[index].new_page_pointer =  Page3;
	}
	
	//Schaltet Ausgnge am AVR je nach Headerinhalt
	//oder Startseite
	if((!http_entry[index].first_switch)&&(http_entry[index].new_page_pointer == Page1))
	{
		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED1=1\0")!=0){ 
			PORTA ^= (1<<PA0);
			http_entry[index].first_switch = 1;
		}
		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED2=2\0")!=0){
			PORTA ^= (1<<PA1);
			http_entry[index].first_switch = 1;
		}
		if (strcasestr((char*)&eth_buffer[TCP_DATA_START_VAR],"LED3=3\0")!=0){
			PORTA ^= (1<<PA2);
			http_entry[index].first_switch = 1;
		}
	}

	//Wurde das Ende vom Header nicht erreicht
	//kommen noch weitere Stcke vom Header!
	if (*http_entry[index].hdr_end_pointer != 0)
	{
		//Der Empfang wird Quitiert und es wird auf weiteres Headerstck gewartet
		tcp_entry[index].status =  ACK_FLAG;
		create_new_tcp_packet(0,index);
		//Warten auf weitere Headerpackete
		tcp_entry[index].app_status = 1;
		return;
	}	
	
	//Wurde das Passwort in den ganzen Headerpacketen gefunden?
	//Wenn nicht dann ausfhren und Passwort anfordern!
	if((!http_entry[index].http_auth) && tcp_entry[index].status&PSH_FLAG)
	{	
		//HTTP_AUTH_Header senden!
		http_entry[index].new_page_pointer = Page0;
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_header3,HTTP_AUTH_HEADER_LEN);
		tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
		create_new_tcp_packet(129,index);
		return;
	}
	
	tcp_entry[index].app_status = 2;
	//Seiten Header wird gesendet
	if(http_entry[index].http_header_type == 1)
	{
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_header2,HTTP_OK_HEADER_LEN);
	}
	else
	{
		memcpy_P((char*)&eth_buffer[TCP_DATA_START_VAR],http_header1,HTTP_OK_HEADER_LEN);
	}
	tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
	create_new_tcp_packet(87,index);
	return;	
}

//----------------------------------------------------------------------------
//Daten Packete an Client schicken
void httpd_data_send (unsigned char index)
{		
	//Passwort wurde im Header nicht gefunden
	if(!http_entry[index].http_auth)
	{
		http_entry[index].new_page_pointer = Page0;
	}
	
	unsigned int a;
	unsigned char str_len;
	char var_conversion_buffer[10];
	
	//kein Packet empfangen Retransmission des alten Packetes
	if (tcp_entry[index].status == 0) 
	{
		http_entry[index].new_page_pointer = http_entry[index].old_page_pointer;
	}
	http_entry[index].old_page_pointer = http_entry[index].new_page_pointer;

	for (a = 0;a<(MTU_SIZE-(TCP_DATA_START)-150);a++)
	{
		unsigned char b;
		b = pgm_read_byte(http_entry[index].new_page_pointer++);
		eth_buffer[TCP_DATA_START + a] = b;
		
		//Mssen Variablen ins Packet eingesetzt werden? ===> %VA@00 bis %VA@09
		if (strncasecmp_P("%VA@",http_entry[index].new_page_pointer-1,4)==0)
		{
			
			b = (pgm_read_byte(http_entry[index].new_page_pointer+3)-48)*10;
			b +=(pgm_read_byte(http_entry[index].new_page_pointer+4)-48);	
			itoa (var_array[b],var_conversion_buffer,10);

			str_len = strnlen(var_conversion_buffer,MAX_VAR_ARRAY);
			memmove(&eth_buffer[TCP_DATA_START+a],var_conversion_buffer,str_len);
			a = a + (str_len-1);
			http_entry[index].new_page_pointer=http_entry[index].new_page_pointer+5;
		}

		//wurde das Ende des Packetes erreicht?
		//Verbindung TCP Port kann beim nchsten ACK geschlossen werden
		//Schleife wird abgebrochen keine Daten mehr!!
		if (strncasecmp_P("%END",http_entry[index].new_page_pointer,4)==0)
		{	
			tcp_entry[index].app_status = 0xFFFD;
			a++;
			break;
		}
	}
	//Erzeugte Packet kann nun gesendet werden!
	tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
	create_new_tcp_packet(a,index);
	return;
}


